/*-------------------------------------------------------------------*/
/* LS.C         Pasi Frnti                                          */
/*                                                                   */
/* Local search routines                                             */
/*   - Random Swap                                                   */
/*   - Deterministic Swap (not yet implemented)                      */
/*   - Including functions for unknown no. of clusters               */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "LS"
#define VersionNumber  "Version 0.12"
#define LastUpdated    "4.5.2000" /* iak */

/* ----------------------------------------------------------------- */

#include <float.h>
#include <stdio.h>
#include <time.h>

#include "ls.h"
#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "cef.h"
#include <string.h>

#include "traceprg.h"

/* #include "cbi.h" */

/* ------------------ Parameters for the LS module ----------------- */

static struct { int QuietLevel;
                int SwappingMethod;
                int GLAIterations;
                int LSIterations;
                int MinBookSize;
                int MaxBookSize;
                ERRORFTYPE ErrorFunction;
                DISTANCETYPE DistanceFunction;
              } ModuleParameters;

#define Value(x)      (ModuleParameters.x)

/*--------------------- Enums for GLA and LS -------------------------*/

/* These enum definitions are from GLA and LS modules. */
enum { NoSwap, RandomSwap, DetermSwap, TabuSwap };
/* enum { GLAFullSearch, GLALocal, GLAWu, GLATIE, GLALocalTIE,
       GLAMPS, GLALocalMPS };
   enum { PDSoff, PDSon };
   enum { FirstVector, CurrentVector };
   enum { Centroid, NearestToCentroid }; */

/*--------------------------- Tabu list ------------------------------*/

#define  TABULISTSIZE   10
int*     Tabulist;
int      Tabupointer;


/*---------------------------  M i s c	------------------------------*/

#define  fpositive(a)   ((a) > FLT_EPSILON)



/*========================= Printing routines ========================*/


static long SetWatch(long* watch)
{
  time_t tmp;
  return( *watch = (long)time(&tmp) );
}


/*-------------------------------------------------------------------*/


static void PrintTime(long watch)
{
  time_t tmp;
  long t = (long)time(&tmp) - watch;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  PrintMessage("  %li:%02li:%02li\n", h, m, s);
}


/*-------------------------------------------------------------------*/


static void PrintProgress(CODEBOOK*   CB,
                          int         iter,
                          double      eb,
                          double      ec,
						  int		  improve,
                          long        watch)
{
  double  Best,Curr;
  Best = PrintableError(eb, CB);
  Curr = PrintableError(ec, CB);
  switch( Value(QuietLevel) )
	{
	case 0:  return;
    case 1:  PrintMessage("%3i: %9.4f \n", iter, Best);
             break;
    case 2:  PrintMessage("%3i: %9.4f  ", iter, Best);
             PrintTime(watch);
             break;
    default: PrintMessage("%3i: best=%9.4f  curr=%9.4f  (%c)",
			   iter, Best, Curr, (improve ? '+' : '-') );
             PrintTime(watch);
             break;

	}
#if !(defined(WIN32) && !defined(BATCH))
  fflush(stdout);
#endif
}


/*-------------------------------------------------------------------*/


static void PrintResult(CODEBOOK*   CB,
                        int         iter,
                        double      e,
                        long        watch)
{
  double  Error;
  Error = PrintableError(e, CB);
  if( Value(QuietLevel) >= 1 ) printf("%9.4f  ", Error);
  if( Value(QuietLevel) >= 1 ) PrintTime(watch);
#if !(defined(WIN32) && !defined(BATCH))
  fflush(stdout);
#endif
}


/*======================= Solution handling =========================*/


static int FreqToIndex(TRAININGSET* TS, int sum)
{
  int i=0,l=0;

  while (l<sum)  l += VectorFreq(TS, i++);
  return (i-1);
}


/*-------------------------------------------------------------------*/


static void GenerateRandomVector(TRAININGSET* TS, CODEBOOK* CB, int vec)
{
  int     i;
  YESNO   found=NO;
  int     t;
  int     index;

  while(!found)
     {
	 found=YES;
     t=irand(1,TotalFreq(TS));
     index=FreqToIndex(TS, t);
	 for (i=0; i<BookSize(CB); i++)
		if (EqualVectors(Vector(TS,index), Vector(CB,i), VectorSize(CB)))
           {
		   found=NO; break;
           }
     }
  CopyVector(Vector(TS,index), Vector(CB,vec), VectorSize(CB));
  VectorFreq(CB,vec)=1;
}


/*-------------------------------------------------------------------*/


static void GenerateRandomCodebook(TRAININGSET* TS, CODEBOOK* CB)
{
  int i;

  for (i=0; i<BookSize(CB); i++)
      {
      GenerateRandomVector(TS,CB,i);
      }
}


/*========================== Initialize =============================*/


void SetLSParameters(int QuietLevel,
                     int SwappingMethod,
                     int GLAIterations,
                     int LSIterations,
                     int MinBookSize,
					 int MaxBookSize,
                     ERRORFTYPE errorf)
{
  Value(QuietLevel)       = QuietLevel;
  Value(SwappingMethod)   = SwappingMethod;
  Value(GLAIterations)    = GLAIterations;
  Value(LSIterations)     = LSIterations;
  Value(MinBookSize)      = MinBookSize;
  Value(MaxBookSize)      = MaxBookSize;
  Value(ErrorFunction)    = errorf;
  Value(DistanceFunction) = DistType(errorf);
}


/*-------------------------------------------------------------------*/


static void InitializeSolution(TRAININGSET*   TS,
							   CODEBOOK*	  CB,
							   PARTITIONING*  P,
							   YESNO		  ImproveExisting)
{
  if( BookSize(TS) < Value(MaxBookSize) )
     {
     ErrorMessage("LS: Training set size (%i) < requested codebook size (%i).\n",
                   BookSize(TS), Value(MaxBookSize));
     }

  if(!ImproveExisting)  /* Create random solution */
	 {
	 ClearGenerationMethod(CB);
     AddGenerationMethod(CB, "Random");
     BookSize(CB) = Value(MaxBookSize);       /* Are these two needed?? */
     PartitionCount(P) = Value(MaxBookSize);  /* Are these two needed?? */
	 GenerateRandomCodebook(TS, CB);
	 GenerateOptimalPartitioning(TS, CB, P);
	 }

  if (!strstr(CB->GenerationMethod, "+LS "))
	 {
	 AddGenerationMethod(CB, "+LS ");
     AddGenerationMethod(CB, VersionNumber);
	 }

}


/*====================== Tabu list management ========================*/


static void InitializeTabuList(void)
{
  int i;

  Tabulist    = (int*) allocate( TABULISTSIZE * sizeof(int) );
  Tabupointer = 0;
  for( i=0; i<TABULISTSIZE; i++ ) Tabulist[i]=0;
}


/*-------------------------------------------------------------------*/


static void AddToTabulist(int x)
{
  Tabulist[Tabupointer++] = x;
  if(Tabupointer==TABULISTSIZE) Tabupointer=0;
}


/*--------------------------------------------------------------------*/


static YESNO IsTabu(int x)
{
  int i;

  for(i=0; i<TABULISTSIZE; i++)
	 if( Tabulist[i]==x ) return(YES);

  return(NO);
}


/*--------------------------------------------------------------------*/


static void PrintTabulist(void)
{
  int i,p;

  PrintMessage("Tabulist: ");
  p=Tabupointer;
  for(i=0; i<TABULISTSIZE; i++)
	 {
     PrintMessage("%3i ",Tabulist[p++]);
	 if(p==TABULISTSIZE) p=0;
	 }
  PrintMessage("\n");
}


/*======================= Reduced search GLA =========================*/


static void DetectChangedCodeVectors(CODEBOOK*      CBprev,
                                     CODEBOOK*      CB,
                                     int*           Changed,
                                     int*           ChangeList,
                                     int*           newcount)
{
  int i;

  (*newcount) = 0;
  for( i=0; i<BookSize(CB); i++ )
    {
    if(!(EqualVectors(Vector(CB,i), Vector(CBprev,i), VectorSize(CB))))
	   {
       Changed[i]=YES; ChangeList[*newcount]=i; (*newcount)++;
	   }
    else
       {
       Changed[i] = NO;
       }
    }
}

/*--------------------------------------------------------------------*/


static int FindNearestVectorInSet(BOOKNODE*     v,
                                  CODEBOOK*     CB,
                                  int           guess,
                                  int*          ChangeList,
                                  int           newcount)
{
  int    i,j;
  int    minindex = guess;
  llong  e,minerror;

  minerror = VectorDistance(Vector(CB,guess), v->vector, VectorSize(CB), MAXLLONG, EUCLIDEANSQ);
  for( i=0; i<newcount; i++ )
     {
     j = ChangeList[i];
     e = VectorDistance(Vector(CB,j), v->vector, VectorSize(CB), minerror, EUCLIDEANSQ);
     if( e<minerror )
        {
        minerror = e;
        minindex = j;
        if( e==0 ) return(minindex);
        }
     }
  return(minindex);
}


/*--------------------------------------------------------------------*/


static void ReducedSearchPartition(TRAININGSET*   TS,
								   CODEBOOK*	  CB,
								   PARTITIONING*  P,
                                   int*           Changed,
                                   int*           ChangeList,
                                   int            newcount)
{
  llong  error;
  int    i, nearest;

  for( i=0; i<BookSize(TS); i++ )
    {
    if( Changed[Map(P,i)] )
       {
       nearest = FindNearestVector(&Node(TS,i), CB, &error, Map(P,i), EUCLIDEANSQ);
       }
    else
       {
       nearest = FindNearestVectorInSet(&Node(TS,i), CB, Map(P,i), ChangeList, newcount);
       }
    if(nearest!=Map(P,i)) ChangePartition(TS,P,nearest,i);
    }
}



/*--------------------------------------------------------------------*/


static void IterateByGLA(TRAININGSET*   TS,
						 CODEBOOK*		CBprev,
						 CODEBOOK*		CB,
                         PARTITIONING*  P,
                         int*           Changed,
                         int*           ChangeList,
                         int            iterations,
                         long           watch,
						 int			EvaluationFunction,
						 int			MovingMethod)
{
  int  newcount=1;
  int  round=0;

  while( round<iterations && newcount>0 )
	{
	CopyCodebook(CB,CBprev); /* This was missing...? */
	GenerateOptimalCodebook(TS, CB, P);
	SaveTraceData(0, 0, CB, 4 + round, "gla_cb");
	FillEmptyPartitions(TS, CB, P);
	OptimizeCentroidPositions(TS, CB, P, EvaluationFunction, MovingMethod);
	SaveTraceData(0, 0, CB, 4 + round, "gla_opt");
    DetectChangedCodeVectors(CBprev, CB, Changed, ChangeList, &newcount);
    ReducedSearchPartition(TS, CB, P, Changed, ChangeList, newcount);
	SaveTraceData(TS, P, 0, 4 + round, "gla_part");
    round++;
    if( Value(QuietLevel)>=5 ) { PrintMessage("GLA(%i):  ",round); PrintTime(watch); }
    }
}


/*======================= Generate solutions =========================*/


static void MakeRandomSwap(TRAININGSET*   TS,
                           CODEBOOK*      CB,
                           PARTITIONING*  P)
{
  int c;

  c = irand(0,BookSize(CB)-1);
  GenerateRandomVector(TS, CB, c);
  SaveTraceData(TS, P, CB, 1, "random");
  LocalRepartitioningGeneral(TS, CB, P, c, EUCLIDEANSQ);
  SaveTraceData(TS, P, CB, 2, "repar1");
  RepartitionDueToNewVectorGeneral(TS, CB, P, c, EUCLIDEANSQ);
  SaveTraceData(TS, P, CB, 3, "repar2");
}
/* Alternative swapping without considering frequencies. */
/*
  c = irand(0,BookSize(CB)-1);
  t = irand(1,BookSize(TS)-1);
  CopyVector(Vector(TS,t), Vector(CB,c), VectorSize(CB));
  VectorFreq(CB,c)=1;
  LocalRepartitioningGeneral(TS, CB, P, c, EUCLIDEANSQ);
  RepartitionDueToNewVectorGeneral(TS, CB, P, c, EUCLIDEANSQ);
*/



/*--------------------------------------------------------------------*/


static void MakeDeterministicSwap(TRAININGSET*	 TS,
                                  CODEBOOK*      CB,
                                  PARTITIONING*  P)
{
/* Idea is to remove the cluster with least increase in distortion,
   and replace it either by random vector, or optimize the location
   of the new vector near cluster with large distortion. Implementation
   details not yet defined. */

/* Not yet implemented */
}


/*--------------------------------------------------------------------*/


static void MakeTabuSwap(TRAININGSET*	TS,
                         CODEBOOK*      CB,
                         PARTITIONING*  P)
{
/* This implementation is merely the Random Swap with addition
   of putting the removed vector index into tabu list.
   The idea is to use the Deterministic Swap with Tabu list,
   but as it is not yet implemented, this is here instead. */

  int c;

  c = irand(0,BookSize(CB)-1);
  while(IsTabu(c)) c=irand(0,BookSize(CB)-1);
  AddToTabulist(c);
  if( Value(QuietLevel)>=5 ) PrintTabulist();
  GenerateRandomVector(TS, CB, c);
  LocalRepartitioningGeneral(TS, CB, P, c, EUCLIDEANSQ);
  RepartitionDueToNewVectorGeneral(TS, CB, P, c, EUCLIDEANSQ);
  /* GenerateOptimalCodebook(TS, CB, P); */
}


/*--------------------------------------------------------------------*/


static void GenerateNewCandidate(TRAININGSET*   TS,
                                 CODEBOOK*      CB,
                                 PARTITIONING*  P,
								 CODEBOOK*		CBnew,
                                 PARTITIONING*  Pnew,
								 CODEBOOK*		CBtmp,
                                 int*           Changed,
                                 int*           ChangeList,
                                 long           watch,
								 int			EvaluationFunction,
								 int			MovingMethod)
{
/*double e1,e2;*/
/*int    iter; */

  CopyCodebook(CB, CBnew);
  CopyCodebook(CB, CBtmp);
  CopyPartitioning(P, Pnew);
  switch( Value(SwappingMethod) )
	 {
     case NoSwap:     break;
     case RandomSwap: MakeRandomSwap(TS, CBnew, Pnew); break;
     case DetermSwap: MakeDeterministicSwap(TS, CBnew, Pnew); break;
     case TabuSwap:   MakeTabuSwap(TS, CBnew, Pnew); break;
     default:         break;
	 }
  if( Value(QuietLevel)>=5 ) { PrintMessage("Swap: "); PrintTime(watch); }
/* SetGLAParameters(0, GLALocal, PDSon, CurrentVector, Centroid, MSE); */
/* GLA(TS, CBnew, NULL, Pnew, Value(GLAIterations), &e1, &e2, &iter); */
  IterateByGLA(TS, CBtmp, CBnew, Pnew, Changed, ChangeList, 
	  Value(GLAIterations), watch, EvaluationFunction, MovingMethod);
}


/*------------------- Local search iterations ------------------------*/


int IterateByLocalSearch(TRAININGSET*  TS,
						  CODEBOOK* 	CB,
						  PARTITIONING* P,
						  YESNO 		ImproveExisting,
						  double*		FinalError,
						  int			EvaluationFunction,
						  int			MovingMethod)

/* The method iterates previous solution (Improve=YES), for which both
   CB and P must be given. Another option (Improve=NO) the method starts
   from random solution. Both data structures CB and P must anyway been
   created. */
{
  CODEBOOK      CBnew;
  CODEBOOK      CBtmp;
  PARTITIONING  Pnew;
  int*          ChangeList;
  int*          Changed;
  double        Error;
  long			watch;
  int			i,improve;
  int           ReturnValue = EVERYTHING_OK;

  SetClusterCount(BookSize(CB));
  SetWatch(&watch);
  CreateNewCodebook(&CBnew, Value(MaxBookSize), TS);
  CreateNewCodebook(&CBtmp, Value(MaxBookSize), TS);
  CreateNewPartitioning(&Pnew, TS, Value(MaxBookSize));
  InitializeTabuList();
  InitializeSolution(TS, CB, P, ImproveExisting);
  ChangeList = (int*) allocate( Value(MaxBookSize) * sizeof(int) );
  Changed    = (int*) allocate( Value(MaxBookSize) * sizeof(int) );
  (*FinalError) = Error = 
	  EvaluatePartitioning(TS, CB, P, EvaluationFunction);
  /*(*FinalError) = Error = AverageErrorForSolution(TS, CB, P, MSE);*/
  PrintProgress(CB, 0, Error, Error, NO, watch);
  for(i=1; i<=Value(LSIterations); i++)
	 {
	 SetIteration(i);
	 SaveTraceData(TS, P, CB, 0, "best");
     GenerateNewCandidate(TS, CB, P, &CBnew, &Pnew, &CBtmp, 
		 Changed, ChangeList, watch, EvaluationFunction, MovingMethod);
	 Error = EvaluatePartitioning(TS, &CBnew, &Pnew, EvaluationFunction);
     /*Error = AverageErrorForSolution(TS, &CBnew, &Pnew, MSE);*/
	 improve = NO;
	 if( fpositive((*FinalError)-Error) )
        {
		improve = YES;
		/*IterateByGLA(TS, CB, &CBnew, &Pnew, Changed, ChangeList, 64, watch, EvaluationFunction);*/
        CopyCodebook(&CBnew, CB);
        CopyPartitioning(&Pnew, P);
		(*FinalError) = Error = 
			EvaluatePartitioning(TS, CB, P, EvaluationFunction);
        }
	 PrintProgress(&CBnew, i, (*FinalError), Error, improve, watch);
	 if (ExitRequested())
		{
		ReturnValue = EXIT_REQUEST_HONOURED;
		break;
		}
	 }

  PrintResult(&CBnew, i, (*FinalError), watch);
  FreeCodebook(&CBnew);
  FreeCodebook(&CBtmp);
  FreePartitioning(&Pnew);
  deallocate(Tabulist);
  deallocate(ChangeList);
  deallocate(Changed);
  return ReturnValue;
}

/* Version of local search that continues until error is less than limit */
/* Needed to test improvements when using previous solution */

#ifdef TESTING

int TestingIterateByLocalSearch(TRAININGSET*  TS,
						  CODEBOOK* 	CB,
						  PARTITIONING* P,
						  YESNO 		ImproveExisting,
						  double*		FinalError,
						  int EvaluationFunction,
						  int			MovingMethod)

/* The method iterates previous solution (Improve=YES), for which both
   CB and P must be given. Another option (Improve=NO) the method starts
   from random solution. Both data structures CB and P must anyway been
   created. */
{
  CODEBOOK      CBnew;
  CODEBOOK      CBtmp;
  PARTITIONING  Pnew;
  int*          ChangeList;
  int*          Changed;
  double        Error;
  long			watch;
  int			i,improve;
  int           ReturnValue = EVERYTHING_OK;
  double		ErrorLimit;
  char			BestName[512];
  
  SetWatch(&watch);
  CreateNewCodebook(&CBnew, Value(MaxBookSize), TS);
  CreateNewCodebook(&CBtmp, Value(MaxBookSize), TS);
  CreateNewPartitioning(&Pnew, TS, Value(MaxBookSize));
  InitializeTabuList();
  InitializeSolution(TS, CB, P, ImproveExisting);
  ChangeList = (int*) allocate( Value(MaxBookSize) * sizeof(int) );
  Changed    = (int*) allocate( Value(MaxBookSize) * sizeof(int) );
  ErrorLimit = *FinalError * 1.000001;
  (*FinalError) = Error = 
	  EvaluatePartitioning(TS, CB, P, EvaluationFunction);
  /*(*FinalError) = Error = AverageErrorForSolution(TS, CB, P, MSE);*/
  PrintProgress(CB, 0, Error, Error, NO, watch);
  i = 0;
  sprintf(BestName, "test_%i_%i.cb", BookSize(CB), i);
  WriteCodebook(BestName, CB, 1);
  while (*FinalError > ErrorLimit)
	 {
	 i++;
     GenerateNewCandidate(TS, CB, P, &CBnew, &Pnew, &CBtmp,
		 Changed, ChangeList, watch, EvaluationFunction);
	 Error = EvaluatePartitioning(TS, &CBnew, &Pnew, EvaluationFunction);
     /*Error = AverageErrorForSolution(TS, &CBnew, &Pnew, MSE);*/
	 improve = NO;
	 if (*FinalError > Error)
        {
		improve = YES;
		/*IterateByGLA(TS, CB, &CBnew, &Pnew, Changed, ChangeList, 64, watch);*/
        CopyCodebook(&CBnew, CB);
        CopyPartitioning(&Pnew, P);
		(*FinalError) = Error = 
			EvaluatePartitioning(TS, CB, P, EvaluationFunction);
		sprintf(BestName, "test_%i_%i.cb", BookSize(CB), i);
		WriteCodebook(BestName, CB, 1);
		fprintf(stderr, "iter: %i %e\n", i, Error);
        }
	 PrintProgress(&CBnew, i, (*FinalError), Error, improve, watch);
	 if (ExitRequested())
		{
		ReturnValue = EXIT_REQUEST_HONOURED;
		break;
		}
	 }
  fprintf(stderr, "%i %i\n", Value(MaxBookSize), i);

  PrintResult(&CBnew, i, (*FinalError), watch);
  FreeCodebook(&CBnew);
  FreeCodebook(&CBtmp);
  FreePartitioning(&Pnew);
  deallocate(Tabulist);
  deallocate(ChangeList);
  deallocate(Changed);
  return ReturnValue;
}

#endif

int BruteForce(TRAININGSET* TSIn, CODEBOOK* CBOut, 
	PARTITIONING* PAOut, int EvaluationFunction, int Minimum, int Maximum,
	int QuietLevel, int SwappingMethod, int GLAIterations, int LSIterations, 
	int OutputGraphData, int InitialSolution, int MovingMethod)
{
	int Interrupt, k;
	double Best, Curr, FinalError;
	CODEBOOK CurrCB;
	PARTITIONING CurrPA;
	SetExitRequestHonored(1);
	if (InitialSolution) {
		SetLSParameters((0 > QuietLevel - 2) ? 0 : (QuietLevel - 2),
			SwappingMethod, GLAIterations,
			LSIterations, BookSize(CBOut), BookSize(CBOut), MSE);
		CreateNewPartitioning(PAOut, TSIn, BookSize(CBOut));
		GenerateOptimalPartitioningGeneral(TSIn, CBOut, PAOut, MSE);
		Interrupt = IterateByLocalSearch(
			TSIn, CBOut, PAOut, YES, &FinalError, EvaluationFunction, MovingMethod);
	} else {
		CreateNewPartitioning(PAOut, TSIn, Minimum);
		CreateNewCodebook(CBOut, Minimum, TSIn);
		GenerateRandomCodebook(TSIn, CBOut);
		GenerateOptimalPartitioningGeneral(TSIn, CBOut, PAOut, MSE);
		GenerateOptimalCodebook(TSIn, CBOut, PAOut);
		Interrupt = 0;
	}
	if (Interrupt) return Interrupt;
	Best = EvaluatePartitioning(TSIn, CBOut, PAOut, EvaluationFunction);
	if (InitialSolution && QuietLevel) PrintMessage(
		"Given solution refined:\n"
		"m = %i, Value = %.6e\n", BookSize(CBOut), Best);
#if defined(SWI)
	OpenProgressWindow();
#elif !(defined(WIN32) && !defined(BATCH))
	if (OutputGraphData) fprintf(stderr, "%i %.6e\n", Minimum, Best);
#endif
	for (k = Minimum; k <= Maximum; k++) {
		CreateNewCodebook(&CurrCB, k, TSIn);
		CreateNewPartitioning(&CurrPA, TSIn, k);
		SetLSParameters(
			(0 > QuietLevel - 2) ? 0 : (QuietLevel - 2),
			SwappingMethod, GLAIterations,
			LSIterations, k, k, MSE);
		Interrupt = IterateByLocalSearch(
			TSIn, &CurrCB, &CurrPA, NO, &FinalError, EvaluationFunction, MovingMethod);
		if (Interrupt) {
			FreeCodebook(&CurrCB);
			FreePartitioning(&CurrPA);
			return Interrupt;
		}
		Curr = EvaluatePartitioning(
			TSIn, &CurrCB, &CurrPA, EvaluationFunction);
#if defined(SWI)
		UpdateProgressBar(k - Minimum, Maximum - Minimum);
#elif !(defined(WIN32) && !defined(BATCH))
		if (OutputGraphData) fprintf(stderr, "%i %.6e\n", k, Curr);
#endif
		if (Curr < Best) {
			FreeCodebook(CBOut);
			CreateNewCodebook(CBOut, k, TSIn);
			CopyCodebook(&CurrCB, CBOut);
			FreePartitioning(PAOut);
			CreateNewPartitioning(PAOut, TSIn, k);
			CopyPartitioning(&CurrPA, PAOut);
			if (QuietLevel) PrintMessage(
				"m = %i, Value = %.6e, delta = %e\n", k, Curr, Best - Curr);
			Best = Curr;
		} else if (QuietLevel > 1) 
			PrintMessage("m = %i, Value = %.6e\n", k, Curr);
		FreeCodebook(&CurrCB);
		FreePartitioning(&CurrPA);
	}
#if defined(SWI)
	CloseProgressWindow();
#endif
	return 0;
}

int Stepwise(TRAININGSET* TSIn, CODEBOOK* CBOut, 
	PARTITIONING* PAOut, int EvaluationFunction, int Minimum, int Maximum,
	int QuietLevel, int SwappingMethod, 
	int GLAIterations, int LSIterations, int LSStepwiseIterations,
	int OutputGraphData, int InitialSolution, int SearchDirectionDecreasing,
	int MovingMethod)
{
	int Interrupt, k, InitialBookSize;
	double Best, Curr;
	CODEBOOK CurrCB;
	PARTITIONING CurrPA;
#ifdef TESTING
	double* Limits, *CurrentLimit;

	CurrentLimit = Limits = allocate(sizeof(double) * (Maximum - Minimum + 1));
	printf("Give %i numbers:\n", (Maximum - Minimum + 1));
	for (k = 0; k < (Maximum - Minimum + 1); k++) {
		Interrupt = scanf("%lf", CurrentLimit++);
		if ((Interrupt == EOF) || !Interrupt) ExitProcessing(FATAL_ERROR);
	}
	printf("Ok.\n");
#endif
	SetExitRequestHonored(1);

	/* generate inital solution or use existing */
	if (InitialSolution) {
		CreateNewPartitioning(PAOut, TSIn, BookSize(CBOut));
		GenerateOptimalPartitioning(TSIn, CBOut, PAOut);
	} else {
		if (SearchDirectionDecreasing) {
			CreateNewCodebook(CBOut, Maximum, TSIn);
			CreateNewPartitioning(PAOut, TSIn, Maximum);
		} else {
			CreateNewCodebook(CBOut, Minimum, TSIn);
			CreateNewPartitioning(PAOut, TSIn, Minimum);
		}
		ClearGenerationMethod(CBOut);
		AddGenerationMethod(CBOut, "Random");
		GenerateRandomCodebook(TSIn, CBOut);
		GenerateOptimalPartitioning(TSIn, CBOut, PAOut);
	}
	InitialBookSize = BookSize(CBOut);

	/* improve the initial solution */
	SetLSParameters((0 > QuietLevel - 2) ? 0 : (QuietLevel - 2),
		SwappingMethod, GLAIterations, LSIterations, 
		BookSize(CBOut), BookSize(CBOut), MSE);
	Interrupt = IterateByLocalSearch(
		TSIn, CBOut, PAOut, YES, &Best, EvaluationFunction, MovingMethod);
	if (Interrupt) {
		FreeCodebook(CBOut);
		FreePartitioning(PAOut);
		return Interrupt;
	}
	/*Best = EvaluatePartitioning(TSIn, CBOut, PAOut, EvaluationFunction);*/
	if (QuietLevel) {
		if (InitialSolution) PrintMessage("Initial solution improved.");
		PrintMessage("m = %i, Value = %.6e\n", BookSize(CBOut), Best);
	}
#if defined(SWI)
	OpenProgressWindow();
#elif !(defined(WIN32) && !defined(BATCH))
	if (OutputGraphData) 
		fprintf(stderr, "%i %.6e\n", BookSize(CBOut), Best);
#endif

	/* check all lesser values */
	CreateNewCodebook(&CurrCB, BookSize(CBOut), TSIn);
	CopyCodebook(CBOut, &CurrCB);
	AddGenerationMethod(&CurrCB, " (+ Random centroid add/delete)");
	for (k = InitialBookSize - 1; k >= Minimum; k--) {
		DecreaseCodebookSize(&CurrCB, k);
		CreateNewPartitioning(&CurrPA, TSIn, k);
		GenerateOptimalPartitioning(TSIn, &CurrCB, &CurrPA);
		SetLSParameters((0 > QuietLevel - 2) ? 0 : (QuietLevel - 2),
			SwappingMethod, GLAIterations, LSStepwiseIterations, 
			k, k, MSE);
#ifdef TESTING
		Curr = Limits[k - Minimum];
		Interrupt = TestingIterateByLocalSearch(
			TSIn, &CurrCB, &CurrPA, YES, &Curr, EvaluationFunction, MovingMethod);
#else
		Interrupt = IterateByLocalSearch(
			TSIn, &CurrCB, &CurrPA, YES, &Curr, EvaluationFunction, MovingMethod);
#endif
		if (Interrupt) {
			FreeCodebook(&CurrCB);
			FreePartitioning(&CurrPA);
			return Interrupt;
		}
		/*Curr = EvaluatePartitioning(
			TSIn, &CurrCB, &CurrPA, EvaluationFunction);*/
#if defined(SWI)
		UpdateProgressBar(k - Minimum, Maximum - Minimum);
#elif !(defined(WIN32) && !defined(BATCH))
		if (OutputGraphData) fprintf(stderr, "%i %.6e\n", k, Curr);
#endif
		if (Curr < Best) {
			FreeCodebook(CBOut);
			CreateNewCodebook(CBOut, k, TSIn);
			CopyCodebook(&CurrCB, CBOut);
			FreePartitioning(PAOut);
			CreateNewPartitioning(PAOut, TSIn, k);
			CopyPartitioning(&CurrPA, PAOut);
			if (QuietLevel) PrintMessage(
				"m = %i, Value = %.6e, delta = %e\n", k, Curr, Best - Curr);
			Best = Curr;
		} else if (QuietLevel > 1) 
			PrintMessage("m = %i, Value = %.6e\n", k, Curr);
		FreePartitioning(&CurrPA);
	}
	/* check all greater values */
	for (k = InitialBookSize + 1; k <= Maximum; k++) {
		IncreaseCodebookSize(&CurrCB, k);
		GenerateRandomVector(TSIn, &CurrCB, k - 1);
		CreateNewPartitioning(&CurrPA, TSIn, k);
		GenerateOptimalPartitioning(TSIn, &CurrCB, &CurrPA);
		SetLSParameters((0 > QuietLevel - 2) ? 0 : (QuietLevel - 2),
			SwappingMethod, GLAIterations, LSStepwiseIterations, 
			k, k, MSE);
#ifdef TESTING
		Curr = Limits[k - Minimum];
		Interrupt = TestingIterateByLocalSearch(
			TSIn, &CurrCB, &CurrPA, YES, &Curr, EvaluationFunction, MovingMethod);
#else
		Interrupt = IterateByLocalSearch(
			TSIn, &CurrCB, &CurrPA, YES, &Curr, EvaluationFunction, MovingMethod);
#endif
		if (Interrupt) {
			FreeCodebook(&CurrCB);
			FreePartitioning(&CurrPA);
			return Interrupt;
		}
		/*Curr = EvaluatePartitioning(
			TSIn, &CurrCB, &CurrPA, EvaluationFunction);*/
#if defined(SWI)
		UpdateProgressBar(k - Minimum, Maximum - Minimum);
#elif !(defined(WIN32) && !defined(BATCH))
		if (OutputGraphData) fprintf(stderr, "%i %.6e\n", k, Curr);
#endif
		if (Curr < Best) {
			FreeCodebook(CBOut);
			CreateNewCodebook(CBOut, k, TSIn);
			CopyCodebook(&CurrCB, CBOut);
			FreePartitioning(PAOut);
			CreateNewPartitioning(PAOut, TSIn, k);
			CopyPartitioning(&CurrPA, PAOut);
			if (QuietLevel) PrintMessage(
				"m = %i, Value = %.6e, delta = %e\n", k, Curr, Best - Curr);
			Best = Curr;
		} else if (QuietLevel > 1) 
			PrintMessage("m = %i, Value = %.6e\n", k, Curr);
		FreePartitioning(&CurrPA);
	}
#ifdef TESTING
	deallocate(Limits);
#endif
	FreeCodebook(&CurrCB);
#if defined(SWI)
	CloseProgressWindow();
#endif
	return 0;
}
