/**********************************************************************

	CEF.C

	Ismo Krkkinen

	Last changed: 4.5.2000

	Contains functions to calculate the fitness of clustering. These
	take into account the number of clusters, so they can be used to
	find the optimal number of clusters in a set.

**********************************************************************/

/* Version 0.08 */

#include "cef.h"
#include "memctrl.h"
#include "owntypes.h"
#include "interfc.h"
#include <math.h>
#include "cb.h"
#include <stdlib.h>
#include <search.h>

/*==================== Davies-Bouldin Index ==========================*/

typedef struct {
	double Dispersion;
	double MaxW2BCS;
	int Contributor;
	int Moved;
} DBIPartition;

static DBIPartition* GetDBIData(PARTITIONING* PA, CODEBOOK* CB, TRAININGSET* TS) {
	int k, n, Vect;
	VECTORTYPE Centroid;
	DBIPartition* P;
	llong temp;
	double W2BCS, Sum;
	P = allocate(sizeof(DBIPartition) * PartitionCount(PA));
	for (k = 0; k < PartitionCount(PA); k++) {
		
		Centroid = Vector(CB, k);
		P[k].Dispersion = 0;
		P[k].MaxW2BCS = 0;
		Sum = 0;
		
		Vect = FirstVector(PA, k);
		while (!EndOfPartition(Vect)) {
			temp = VectorDist(Vector(TS, Vect), Centroid, VectorSize(TS));
			Sum += sqrt((double) temp);
			Vect = NextVector(PA, Vect);
		}
		
		P[k].Dispersion = Sum / (double) CCFreq(PA, k);

		for (n = 0; n < k; n++) {
			W2BCS = P[k].Dispersion + P[n].Dispersion;
			temp = VectorDist(Centroid, Vector(CB, n), VectorSize(TS));
			/* EUCLIDEANSQ returns square of the distance */
			W2BCS /= sqrt((double) temp);
			if (W2BCS > P[k].MaxW2BCS) {
				P[k].MaxW2BCS = W2BCS;
				P[k].Contributor = n;
			}
			if (W2BCS > P[n].MaxW2BCS) {
				P[n].MaxW2BCS = W2BCS;
				P[n].Contributor = k;
			}
		}
	}
	return P;
}

static double DaviesBouldinIndex(PARTITIONING* PA, CODEBOOK* CB, TRAININGSET* TS) {
	int k;
	DBIPartition* P;
	double Result = 0.0;
	P = GetDBIData(PA, CB, TS);
	for (k = 0; k < PartitionCount(PA); k++) Result += P[k].MaxW2BCS;
	deallocate(P);
	return Result / (double) PartitionCount(PA);
}


/*============================ Stochastic Complexity =================*/

double* Sums = 0;
double* Logs = 0;

static void scInitializeSumTable(int Count) {
	int k;
	double sum;
	sum = 0.0;
	Sums = allocate(Count * sizeof(double));
	for (k = 1; k < Count; k++) {
		sum += log10((double) k) / log10(2.0);
		Sums[k] = sum;
	}
}

static void scInitializeLogTable(int Count) {
	int k;
	Logs = allocate(Count * sizeof(double));
	for (k = 1; k < Count; k++) Logs[k] = log10(k) / log10(2.0);
}

static void scDestroyTables() {
	if (Sums) deallocate(Sums);
	Sums = 0;
	if (Logs) deallocate(Logs);
	Logs = 0;
}

static double StochasticComplexityBinary(PARTITIONING* PA, TRAININGSET* TS) {
	int Vect, vc, Length, k, p;
	int* Freqs;
	double Result = 0;

	Length = VectorSize(TS);
	Freqs = allocate(sizeof(int) * Length);
	for (p = 0; p < Length; p++) Freqs[p] = 0;
	scInitializeSumTable(TotalFreq(TS) + PartitionCount(PA));
	
	Result = Sums[TotalFreq(TS) + PartitionCount(PA) - 1];
	Result -= Sums[PartitionCount(PA) - 1];
	
	for (k = 0; k < PartitionCount(PA); k++) {
		vc = CCFreq(PA, k);
		Result += -Sums[vc] + Sums[vc + 1] * VectorSize(TS);
		
		Vect = FirstVector(PA, k);
		while (!EndOfPartition(Vect)) {
			for (p = 0; p < Length; p++)
				Freqs[p] += VectorFreq(TS, Vect) *
					(VectorScalar(TS, Vect, p) ? 1 : 0);
			Vect = NextVector(PA, Vect);
		}
		
		for (p = 0; p < Length; p++) {
			Result -= Sums[Freqs[p]] + Sums[vc - Freqs[p]];
			Freqs[p] = 0;
		}
	}
	scDestroyTables();
	deallocate(Freqs);
	
	return Result / (double) TotalFreq(TS);
}

static double XieBeni(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
	int k, n;
	llong Nearest = MAXLLONG;
	llong Dist;
	double mse = EvaluatePartitioning(TS, CB, PA, 0);
	for (k = 0; k < BookSize(CB) - 1; k++) for (n = k + 1; n < BookSize(CB); n++) {
		Dist = VectorDist(
			Vector(CB, k), Vector(CB, n), VectorSize(CB));
		if (Dist < Nearest) Nearest = Dist;
	}
	if (Nearest) return mse / ((double) Nearest / VectorSize(CB));
	else return 1000000000.0;
}

static double Kwon(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
	int k, n;
	llong Nearest = MAXLLONG;
	llong Dist;
	llong CentroidMSE = 0;
	llong* temp;
	VECTORTYPE v;
	double se = BookSize(TS) * EvaluatePartitioning(TS, CB, PA, 0);
	for (k = 0; k < BookSize(CB) - 1; k++) for (n = k + 1; n < BookSize(CB); n++) {
		Dist = VectorDist(
			Vector(CB, k), Vector(CB, n), VectorSize(CB));
		if (Dist < Nearest) Nearest = Dist;
	}
	temp = allocate(sizeof(llong) * VectorSize(TS));
	for (n = 0; n < VectorSize(TS); n++) temp[n] = 0;
	for (k = 0; k < BookSize(TS); k++) for (n = 0; n < VectorSize(TS); n++)
		temp[n] += VectorScalar(TS, k, n);
	for (n = 0; n < VectorSize(TS); n++) temp[n] /= BookSize(TS);
	v = CreateEmptyVector(BookSize(TS));
	for (n = 0; n < BookSize(TS); n++) v[n] = (int) temp[n];
	deallocate(temp);
	for (k = 0; k < BookSize(CB); k++) CentroidMSE += VectorDist(
		v, Vector(CB, k), VectorSize(CB));
	FreeVector(v);
	CentroidMSE /= (BookSize(CB) * VectorSize(CB));
	if (Nearest) return (se + (double) CentroidMSE)
		/ ((double) Nearest / (double) VectorSize(CB));
	else return 1000000000.0;
}

static void MeanOfTS(TRAININGSET* TS, VECTORTYPE Result) {
	llong temp[100];
	int n, k;
	for (k = 0; k < VectorSize(TS); k++) temp[k] = 0;
	for (n = 0; n < BookSize(TS); n++) for (k = 0; k < VectorSize(TS); k++)
		temp[k] += VectorScalar(TS, n, k);
	for (k = 0; k < VectorSize(TS); k++) 
		Result[k] = (int)(temp[k] / BookSize(TS));
}

static double FTest(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
	int k;
	llong QB = 0;
	double NMSE;
	VECTORELEMENT v[100];
	if ((BookSize(CB) <= 1) && (BookSize(TS) == BookSize(CB)))
		return 1000000000.0;
	MeanOfTS(TS, v);
	for (k = 0; k < BookSize(CB); k++) QB += VectorFreq(CB, k) 
		* VectorDist(v, Vector(CB, k), VectorSize(CB));
	QB /= VectorSize(TS);
	NMSE = BookSize(TS) * EvaluatePartitioning(TS, CB, PA, 0);
	NMSE /= BookSize(TS) - BookSize(CB);
	return 1.0 / ((QB / (double)(BookSize(CB) - 1)) / NMSE);
}

static double W[10000];

static void WMatrix(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
	int i, k, n;
	double v[100];
	int Size = VectorSize(TS);
	if ((BookSize(CB) < 2) || (BookSize(TS) == BookSize(CB))) {
		for (k = 0; k < Size * Size; k++) W[k] = 1000000.0;
		return;
	}
	for (k = 0; k < Size * Size; k++) W[k] = 0;
	for (n = 0; n < BookSize(TS); n++) {
		for (k = 0; k < Size; k++) v[k] 
			= (VectorScalar(TS, n, k)
			- VectorScalar(CB, Map(PA, n), k)) /
			(double) TS->MaxValue;
		for (i = 0; i < Size; i++) for (k = 0; k < Size; k++)
			W[i * Size + k] += v[i] * v[k];
	}
	/*for (k = 0; k < Size * Size; k++)
		W[k] /= (BookSize(TS) - BookSize(CB));*/
}

static double B[10000];

static void BMatrix(TRAININGSET* TS, CODEBOOK* CB) {
	int i, k, n;
	VECTORELEMENT A[100];
	double v[100];
	int Size = VectorSize(TS);
	MeanOfTS(TS, A);
	for (k = 0; k < Size * Size; k++) B[k] = 0;
	for (n = 0; n < BookSize(CB); n++) {
		for (k = 0; k < Size; k++) v[k] = VectorFreq(CB, n) *
			(VectorScalar(CB, n, k) - A[k]) / (double) TS->MaxValue;
		for (i = 0; i < Size; i++) for (k = 0; k < Size; k++)
			B[i * Size + k] += v[i] * v[k];
	}
	/*for (k = 0; k < Size * Size; k++)
		B[k] /= BookSize(TS);*/
}

static double trace(double* Matrix, int Size) {
	int k;
	double Result = 0;
	for (k = 0; k < Size; k++)
		Result += Matrix[k + k * Size] * Matrix[k + k * Size];
	return Result;
}

static double CalinskiHarabasz(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA)
{
	double tW, tB;
	WMatrix(TS, CB, PA);
	BMatrix(TS, CB);
	tW = trace(W, VectorSize(TS));
	tB = trace(B, VectorSize(TS));
	if ((BookSize(CB) > 1) && (BookSize(TS) != BookSize(CB)))
		return (tW / (BookSize(TS) - BookSize(CB)))
			/ (tB / (BookSize(CB) - 1));
	return 1000000000.0;
}

static double LUDecomposition(double* Matrix, int Size) {
	double dd, d = 1.0;
	int i, imax, j, k;
	double big, dum, sum, temp;
	double* vv;
	vv = allocate(sizeof(double) * Size);
	for (i = 0; i < Size; i++) {
		big = 0;
		for (j = 0; j < Size; j++) if (
			(temp = (Matrix[i * Size + j] < 0) ?
				-Matrix[i * Size + j] :
				Matrix[i * Size + j]) > big
			) big = temp;
		if (big == 0.0) {
			for (k = 0; k < Size * Size; k++) Matrix[k] = 0;
			return 1.0;
		}
		vv[i] = 1.0 / big;
	}
	for (j = 0; j < Size; j++) {
		for (i = 0; i < j; i++) {
			sum = Matrix[i * Size + j];
			for (k = 0; k < i; k++) sum -=
				Matrix[i * Size + k] * Matrix[k * Size + j];
			Matrix[i * Size + j] = sum;
		}
		big = 0;
		for (i = j; i < Size; i++) {
			sum = Matrix[i * Size + j];
			for (k = 0; k < j; k++) sum -=
				Matrix[i * Size + k] * Matrix[k * Size + j];
			Matrix[i * Size + j] = sum;
			if ((dum = vv[i] * (sum < 0) ? -sum : sum) >= big) {
				big = dum;
				imax = i;
			}
		}
		if (j != imax) {
			for (k = 0; k < Size; k++) {
				dum = Matrix[imax * Size + k];
				Matrix[imax * Size + k] = Matrix[j * Size + k];
				Matrix[j * Size + k] = dum;
			}
			d = -d;
			vv[imax] = vv[j];
		}
		if (j != Size) {
			dd = 1.0 / Matrix[j * Size + j];
			for (i = j + 1; i < Size; i++)
				Matrix[i * Size + j] *= dd;
		}
	}
	deallocate(vv);
	return d;
}

static double det(double* Matrix, int Size) {
	LUDecomposition(Matrix, Size);
	return trace(Matrix, Size);
}

static double Marriott(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
	WMatrix(TS, CB, PA);
	return BookSize(CB) * BookSize(CB) * det(W, VectorSize(TS));
}

static double XuMSE(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
	int k;
	double temp, power = 1.0;
	temp = EvaluatePartitioning(TS, CB, PA, 0) * VectorSize(TS);
	for (k = 0; k < VectorSize(TS); k++) power *= temp;
	power = sqrt(power);
	return power * BookSize(CB);
}

double EvaluatePartitioning(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
	int EvaluationFunction)
{
	switch (EvaluationFunction) {
	case 0: return AverageErrorForSolution(TS, CB, PA, MSE);
	case 1: return StochasticComplexityBinary(PA, TS);
	case 2: return DaviesBouldinIndex(PA, CB, TS);
	case 3: return XieBeni(TS, CB, PA);
	case 4: return Kwon(TS, CB, PA);
	case 5: return FTest(TS, CB, PA);
	case 6: return CalinskiHarabasz(TS, CB, PA);
	case 7: return Marriott(TS, CB, PA);
	case 8: return XuMSE(TS, CB, PA);
	default: ErrorMessage(
		"cef.c::EvaluatePartitioning: Unknown evaluation function (%i)",
		EvaluationFunction);
		ExitProcessing(RECOVERABLE_ERROR_UNCONTROLLED_EXIT);
	}
	return 1000000000.0;
}



static void MoveForDBIPairwise(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
	int k, n;
	double Magn, Length;
	double* Vect;
	DBIPartition* P = GetDBIData(PA, CB, TS);
	for (k = 0; k < BookSize(CB); k++) P[k].Moved = 0;
	Vect = allocate(sizeof(double) * VectorSize(TS));
	for (k = 0; k < BookSize(CB); k++) {
		Magn = -1 + sqrt(1 + P[k].Dispersion + P[P[k].Contributor].Dispersion);
		for (n = 0; n < VectorSize(TS); n++) Vect[n] = 
			VectorScalar(CB, k, n) - VectorScalar(CB, P[k].Contributor, n);
		Length = 0;
		for (n = 0; n < VectorSize(TS); n++) Length += Vect[n] * Vect[n];
		Length = sqrt(Length);
		for (n = 0; n < VectorSize(TS); n++) Vect[n] = Magn * Vect[n] / Length;
		for (n = 0; n < VectorSize(TS); n++) 
			VectorScalar(CB, k, n) += Vect[n];
		P[k].Moved = 1;
	}
	deallocate(Vect);
	deallocate(P);
}

typedef struct {
	int group;
	int refcount;
	int listbegin;
} grl;

typedef struct {
	int next;
	int group;
} referee;

/*
static int grl_comp( const void *arg1, const void *arg2 ) {
	if (((grl*) arg1)->group > ((grl*) arg2)->group) return -1;
	if (((grl*) arg1)->group < ((grl*) arg2)->group) return 1;
	return 0;
}
*/

static void MoveForDBIAll(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA) {
	int k, n, g, Selected, freeposition;
	double Magn, Length;
	double* Vect, *Total;
	grl* groupdata;
	referee* list;
	CODEBOOK DeltaCB;
	DBIPartition* P = GetDBIData(PA, CB, TS);
	list = allocate(sizeof(referee) * (2 * BookSize(CB) + 1));
	groupdata = allocate(sizeof(grl) * BookSize(CB));
	CreateNewCodebook(&DeltaCB, BookSize(CB), TS);
	for (k = 0; k < 2 * BookSize(CB) + 1; k++) list[k].next = 0;
	freeposition = 1;
	for (k = 0; k < BookSize(CB); k++) {
		groupdata[k].group = k;
		groupdata[k].listbegin = freeposition++;
		list[groupdata[k].listbegin].group = P[k].Contributor;
		groupdata[k].refcount = 1;
	}
	for (k = 0; k < BookSize(CB); k++) {
		groupdata[P[k].Contributor].refcount++;
		list[freeposition].next = groupdata[P[k].Contributor].listbegin;
		groupdata[P[k].Contributor].listbegin = freeposition++;
		list[groupdata[P[k].Contributor].listbegin].group = k;
	}
	Vect = allocate(sizeof(double) * VectorSize(TS));
	Total = allocate(sizeof(double) * VectorSize(TS));
	for (k = 0; k < BookSize(CB); k++) {
		n = groupdata[k].listbegin;
		for (n = 0; n < VectorSize(TS); n++) Total[n] = 0;
		g = groupdata[k].listbegin;
		while (g) {
			Selected = list[g].group;
			Magn = -1 + sqrt(1 + P[k].Dispersion + P[Selected].Dispersion);
			for (n = 0; n < VectorSize(TS); n++) Vect[n] = 
				VectorScalar(CB, k, n) - VectorScalar(CB, Selected, n);
			Length = 0;
			for (n = 0; n < VectorSize(TS); n++) Length += Vect[n] * Vect[n];
			Length = sqrt(Length);
			for (n = 0; n < VectorSize(TS); n++) 
				Total[n] += Magn * Vect[n] / Length;
			g = list[g].next;
		}
		for (n = 0; n < VectorSize(TS); n++) 
			VectorScalar(&DeltaCB, k, n) =
				(int)(Total[n] / groupdata[k].refcount);
	}
	for (k = 0; k < BookSize(CB); k++)
		for (n = 0; n < VectorSize(TS); n++)
			VectorScalar(CB, k, n) += VectorScalar(&DeltaCB, k, n);
	FreeCodebook(&DeltaCB);
	deallocate(Total);
	deallocate(Vect);
	deallocate(groupdata);
	deallocate(list);
	deallocate(P);
}


void OptimizeCentroidPositions(TRAININGSET* TS, CODEBOOK* CB,
	PARTITIONING* P, int EvaluationMethod, int MovingMethod)
{
	if (EvaluationMethod == 2) {
		switch (MovingMethod) {
		case 0: break;
		case 1: /* Pairwise */
			MoveForDBIPairwise(TS, CB, P);
			break;
		case 2: /* Averaged */
			MoveForDBIAll(TS, CB, P);
			break;
		default:
			ErrorMessage(
				"cef.c::OptimizeCentroidPositions: "
				"Unknown evaluation function / moving method pair (%i, %i).\n",
				EvaluationMethod, MovingMethod);
			ExitProcessing(-1);
		}
	}
}
