/********************************************************************

	swi_DataClass.c

	Ismo Krkkinen

	Last changed 16.11.1999

	Functions to declare, define and destroy classes of data
	and to add, remove and manipulate data sets belonging to
	a class.

********************************************************************/

#include "swi_DataClass.h"
#include "memctrl.h"
#include "swi_Internal.h"
#include "swi_DataSetTextWindow.h"
#include "swi_SelectedData.h"
#include <process.h>
#include <errno.h>
#include "swi_Requirement.h"
#include "swi_Filter.h"

typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	char* Name;
	int SaveOnExit;
	int UsageCount;
	int DeleteWhenUseless;
	SWI_ID FileFilter;
	SWI_INT_PPVOID_PCHAR_INT_PSWI_ID Load;
	SWI_INT_PVOID_PCHAR Save;
	SWI_INT_PVOID Destroy;
	SWI_INT_PVOID_PINT LineCount;
	SWI_PCHAR_PVOID_INT Line;
	SWI_ID Req;
	int Icon;
	SWI_ID ViewParams;
} DataClass;


/* data class functions */

SWI_ID swiDCNew(char* ClassName, SWI_ID Requirements, int Icon) {
	DataClass* DC;
	DC = allocate(sizeof(DataClass));
	DC->Type = DATA_CLASS;
	if (ClassName) {
		DC->Name = allocate(strlen(ClassName) + 1);
		strcpy(DC->Name, ClassName);
	} else DC->Name = 0;
	DC->SaveOnExit = 1;
	DC->UsageCount = 0;
	DC->DeleteWhenUseless = 0;
	DC->FileFilter = 0;
	DC->Load = 0;
	DC->Save = 0;
	DC->Destroy = 0;
	DC->LineCount = 0;
	DC->Line = 0;
	DC->Req = Requirements;
	DC->Icon = Icon;
	DC->ViewParams = 0;
	swiSInsert(DC);
	return DC->ID;
}

void swiDCDelete(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) {
		if (DC->UsageCount > 0) DC->DeleteWhenUseless = 1;
		else {
			swiFLUnuse(DC->FileFilter);
			swiSRemove(DataClassID, DATA_CLASS);
			if (DC->Name) deallocate(DC->Name);
			deallocate(DC);
		}
	}
}

void swiDCLoadFunc(SWI_ID DataClassID, 
	SWI_INT_PPVOID_PCHAR_INT_PSWI_ID LoadFunc)
{
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) DC->Load = LoadFunc;
}

void swiDCSaveFunc(SWI_ID DataClassID, SWI_INT_PVOID_PCHAR SaveFunc) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) DC->Save = SaveFunc;
}

void swiDCDestroyFunc(SWI_ID DataClassID, SWI_INT_PVOID DestroyFunc) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) DC->Destroy = DestroyFunc;
}

void swiDCTextFunc(SWI_ID DataClassID, 
	SWI_INT_PVOID_PINT LineCountFunc, SWI_PCHAR_PVOID_INT LineFunc)
{
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) {
		DC->LineCount = LineCountFunc;
		DC->Line = LineFunc;
	}
}

void swiDCAskSaveOnExit(SWI_ID DataClassID, int YesOrNo) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) DC->SaveOnExit = YesOrNo;
}

int swiDCSaveOnExit(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->SaveOnExit;
	return 0;
}

char* swiDCDescription(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->Name;
	return 0;
}

int swiDCIcon(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->Icon;
	return 0;
}

SWI_ID swiDCFilter(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->FileFilter;
	return 0;
}

void swiDCSetFilter(SWI_ID DataClassID, SWI_ID FilterID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) {
		if (DC->FileFilter) swiFLUnuse(DC->FileFilter);
		DC->FileFilter = FilterID;
		swiFLUse(FilterID);
	}
}

SWI_ID swiDCRequirement(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->Req;
	return 0;
}

void swiDCUse(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) DC->UsageCount++;
}

void swiDCUnuse(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) {
		DC->UsageCount--;
		if ((DC->UsageCount == 0) && DC->DeleteWhenUseless)
			swiDCDelete(DataClassID);
	}
}

SWI_ID swiDCViewParameters(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->ViewParams;
	return 0;
}

void swiDCSetViewParameters(SWI_ID DataClassID, SWI_ID Params) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) DC->ViewParams = Params;
}

/* data set functions */

typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	SWI_ID ClassID;
	void* External;
	int Enabled;
	int Changed;
	int Saved;
	char* Name;
	int MarkCount;
	int UsageCount;
	int DeleteWhenUseless;
	int Icon;
	HWND TextWindow;
} DataSet;

SWI_ID swiDSNew(SWI_ID DataClassID, void* External, char* SetName) {
	DataSet* DS;
	if (!swiSFind(DataClassID, DATA_SET)) return 0;
	swiDCUse(DataClassID);
	DS = allocate(sizeof(DataSet));
	DS->Type = DATA_SET;
	if (SetName) {
		DS->Name = allocate(strlen(SetName) + 1);
		strcpy(DS->Name, SetName);
	} else DS->Name = 0;
	DS->ClassID = DataClassID;
	DS->External = External;
	DS->Enabled = 1;
	DS->Changed = 1;
	DS->Saved = 0;
	DS->UsageCount = 0;
	DS->DeleteWhenUseless = 0;
	DS->Icon = swiDCIcon(DataClassID);
	DS->TextWindow = 0;
	swiSInsert(DS);
	return DS->ID;
}

void swiDSUnlock(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) {
/*		if (DS->TextWindow) 
			PostMessage(DS->TextWindow, SWI_WM_ENABLED, 1, 0);*/
		DS->Enabled = 1;
	}
}

void swiDSLock(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) {
/*		if (DS->TextWindow) 
			PostMessage(DS->TextWindow, SWI_WM_ENABLED, 0, 0);*/
		DS->Enabled = 0;
	}
}

void swiDSDelete(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) {
		if (DS->UsageCount > 0) DS->DeleteWhenUseless = 1;
		else {
			if (DS->TextWindow) swiDSTWDestroy(DS->TextWindow);
			swiDSDestroy(DataID);
			swiDCUnuse(DS->ClassID);
			swiSRemove(DataID, DATA_SET);
			if (DS->Name) deallocate(DS->Name);
			deallocate(DS);
		}
	}
}

int swiDSLocked(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) return !DS->Enabled;
	return 0;
}

void swiDSChanged(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) {
		DS->Changed = 1;
		DS->Saved = 0;
	}
}

int swiDSHasChanged(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) return DS->Changed;
	return 0;
}

int swiDSHasBeenSaved(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) return DS->Saved;
	return 1;
}

SWI_ID swiDSGetClassID(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) return DS->ClassID;
	return 0;
}

char* swiDSName(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) return DS->Name;
	return 0;
}

int swiDSIcon(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) return DS->Icon;
	return 0;
}

void swiDSSetName(SWI_ID DataID, char* NewName) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS && (DS->Name != NewName)) {
		if (DS->Name) deallocate(DS->Name);
		if (NewName) {
			DS->Name = allocate(strlen(NewName) + 1);
			DS->Saved = 0;
			strcpy(DS->Name, NewName);
		} else DS->Name = 0;
		if (DS->TextWindow)
			PostMessage(DS->TextWindow, SWI_WM_RENAMED, 0, 0);
		PostMessage(swiMainWindowHandle(), 
			SWI_WM_RENAMED, 0, (LPARAM) DataID);
	}
}

void* swiDSExternal(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) return DS->External;
	return 0;
}

void swiDSSetExternal(SWI_ID DataID, void* NewExternal) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) DS->External = NewExternal;
}

void swiDSUse(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) DS->UsageCount++;
}

void swiDSUnuse(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) {
		DS->UsageCount--;
		if ((DS->UsageCount == 0) && DS->DeleteWhenUseless)
			swiDSDelete(DataID);
	}
}

int swiDSCreateTextWindow(SWI_ID DataID) {
	WINDOWPLACEMENT WP;
	DataClass* DC;
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS && !DS->TextWindow) {
		DC = swiSFind(swiDSGetClassID(DataID), DATA_CLASS);
		if (DC->LineCount && DC->Line)
			DS->TextWindow = swiDSTWCreate(DataID);
	} else if (DS && DS->TextWindow) {
		WP.length = sizeof(WINDOWPLACEMENT);
		GetWindowPlacement(DS->TextWindow, &WP);
		if (WP.showCmd == SW_SHOWMINIMIZED)
			ShowWindow(DS->TextWindow, SW_RESTORE);
		SetForegroundWindow(DS->TextWindow);
	} else return 0;
	return 1;
}

void swiDSTextWindowClosed(SWI_ID DataID) {
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS) DS->TextWindow = 0;
}


/* user defined function existence checks */

int swiDCLoadDefined(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->Load != 0;
	return 0;
}

int swiDCSaveDefined(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->Save != 0;
	return 0;
}

int swiDCDestroyDefined(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->Destroy != 0;
	return 0;
}

int swiDCTextViewDefined(SWI_ID DataClassID) {
	DataClass* DC = swiSFind(DataClassID, DATA_CLASS);
	if (DC) return DC->Line != 0 && DC->LineCount != 0;
	return 0;
}


/* user defined function wrappers */

typedef struct {
	SWI_ID ClassID;
	char* Filename;
} LoadTemp;

void __cdecl swiLoadHelper(void* LoadTempPtr) {
	char* foo, *bar;
	SWI_ID DataID, Req, *IDs;
	int Count;
	DataClass* DC;
	DataSet* DS;
	LoadTemp* LT = (LoadTemp*) LoadTempPtr;
	DataID = swiDSNew(LT->ClassID, 0, LT->Filename);
	deallocate(LT->Filename);
	DS = swiSFind(DataID, DATA_SET); 
	DC = swiSFind(LT->ClassID, DATA_CLASS);
	if (DC && (DC->Load != 0)) {
		Req = swiDCRequirement(LT->ClassID);
		Count = swiSDSelectedCount(SWI_SELECTED_ACTIVE);
		if (Count) IDs = allocate(Count * sizeof(SWI_ID));
		else IDs = 0;
		if (swiSDSets(SWI_SELECTED_ACTIVE, Req, Count, IDs)) {
			DS->Changed = 0;
			DS->Saved = 1;
			if (!(*(DC->Load))
				(&(DS->External), swiDSName(DataID), Count, IDs))
				PostMessage(swiMainWindowHandle(),
					SWI_WM_NEW_DATASET, 0, DataID);
			else swiDSDelete(DataID);
		} else {
			swiDSDelete(DataID);
			foo = AppendString(
				"Selected items do not match the requirements for\n",
				swiDCDescription(LT->ClassID));
			bar = AppendString(foo, ".\nThe requirements are:");
			deallocate(foo);
			foo = AppendString(bar, swiRQDescription(Req));
			deallocate(bar);
			swiErrorMessageBox(foo);
			deallocate(foo);
		}
		if (IDs) deallocate(IDs);
		deallocate(LT);
		return;
	}
	deallocate(LT);
	swiDSDelete(DataID);
}


void swiDSLoad(SWI_ID ClassID, char* Filename) {
	unsigned long TH;
	LoadTemp* LT = allocate(sizeof(LoadTemp));
	LT->ClassID = ClassID;
	LT->Filename = Filename;
	TH = _beginthread(&swiLoadHelper, 4096, (void*) LT);
	if (TH == -1) {
		if (errno == EAGAIN) {
			Sleep(2000);
			TH = _beginthread(&swiLoadHelper, 4096, (void*) LT);
			if (TH == -1) swiErrorMessageBox("Too many threads.\n"
				"Failed to create thread to read file, try again later.");
		} else if (errno == EINVAL) swiErrorMessageBox(
			"swi_DataClass.c::swiDSLoad:\n"
			"Invalid arguments to _beginthread, "
			"contact the software vendor.");
		else swiErrorMessageBox("Error creating thread, open failed.");
		deallocate(LT);
	}
}

int swiDSSave(SWI_ID DataID, char* Filename) {
	DataClass* DC;
	int Result;
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS && DS->Enabled) {
		DC = swiSFind(swiDSGetClassID(DataID), DATA_CLASS);
		if (DC && (DC->Save != 0)) {
			Result = ((*(DC->Save))(DS->External, Filename));
			if (!Result) {
				swiDSSetName(DataID, Filename);
				DS->Saved = 1;
				DS->Changed = 0;
			}
			return Result;
		}
	}
	return 0;
}

int swiDSDestroy(SWI_ID DataID) {
	DataClass* DC;
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS && DS->Enabled) {
		DC = swiSFind(swiDSGetClassID(DataID), DATA_CLASS);
		if (DC && DC->Destroy) return (*(DC->Destroy))(DS->External);
	}
	return 0;
}

int swiDSLineCount(SWI_ID DataID, int* ApproximateMaxLineLength) {
	DataClass* DC;
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS && DS->Enabled) {
		DC = swiSFind(swiDSGetClassID(DataID), DATA_CLASS);
		if (DC && DC->LineCount) return (*(DC->LineCount))
			(DS->External, ApproximateMaxLineLength);
	}
	return 0;
}

char* swiDSLine(SWI_ID DataID, int LineNumber) {
	DataClass* DC;
	DataSet* DS = swiSFind(DataID, DATA_SET);
	if (DS && DS->Enabled) {
		DC = swiSFind(swiDSGetClassID(DataID), DATA_CLASS);
		if (DC && DC->Line) return (*(DC->Line))(DS->External, LineNumber);
	}
	return 0;
}
