/********************************************************************

	swi_Filter.c

	Ismo Krkkinen

	Last changed 12.11.1999

	Implements filter that is used with Windows standard dialogs.

	Each filter is an ordered group of description/filterstring -pairs.
	User can select the filter to be used from the drop-down list in
	the Open/Save dialog.

	The strings handled by Length() and Concatenate must consist of
	null-separated strings with two nulls ending the string.

	The first filter specified will be the one offered to the user
	first. It might be wise to add ("All files", "*.*") -pair as the
	last filter.

	Example:
		Description: "Documents"
		FilterString: "*.txt;*.doc"
		DefaultExtension: "txt"

********************************************************************/

#include "swi_Filter.h"
#include "memctrl.h"
#include <string.h>
#include "idxqueue.h"
#include "swi_Storage.h"
#include "swi_DataClass.h"

SWI_ID GlobalOpenFilter = 0;

typedef struct {
	SWI_ID ID;
	SWIDataTypes Type;
	char* FilterString;
	char DefaultExtension[8];
	int UsageCount;
	int DeleteWhenUseless;
} Filter;


static int Length(char* FilterString) {
	int Local, Total = 0;
	char* Curr = FilterString;
	do {
		Local = strlen(Curr) + 1;
		Curr += Local;
		Total += Local;
	} while (*Curr);
	return Total;
}

static char* Concatenate(char* Old, char* Add) {
	char* C, *Curr, *NewCurr;
	int Lo, La, NewLen;
	Lo = Length(Old);
	La = Length(Add);
	NewLen = Lo + La + 1;
	NewCurr = C = allocate(NewLen);
	Curr = Old;
	do {
		strcpy(NewCurr, Curr);
		Lo = strlen(Curr) + 1;
		NewCurr += Lo;
		Curr += Lo;
	} while (*Curr);
	Curr = Add;
	do {
		strcpy(NewCurr, Curr);
		La = strlen(Curr) + 1;
		NewCurr += La;
		Curr += La;
	} while (*Curr);
	*NewCurr = 0;
	return C;
}

void swiFLInitialize() {
	GlobalOpenFilter = swiFLNew();
}

SWI_ID swiFLNew() {
	Filter* F;
	F = allocate(sizeof(Filter));
	F->Type = FILTER;
	F->FilterString = 0;
	F->DefaultExtension[0] = 0;
	F->UsageCount = 0;
	F->DeleteWhenUseless = 0;
	swiSInsert(F);
	return F->ID;
}

void swiFLAddStrings(SWI_ID FilterID, char* Description, char* FilterString) {
	char* NFS, *DF, *DFEnd;
	Filter* F = swiSFind(FilterID, FILTER);
	if (F) {
		DF = allocate(strlen(Description) + 1 + strlen(FilterString) + 2);
		DFEnd = DF + strlen(Description) + 1;
		strcpy(DF, Description);
		strcpy(DFEnd, FilterString);
		DFEnd[strlen(FilterString) + 1] = 0;
		if (F->FilterString) {
			NFS = Concatenate(F->FilterString, DF);
			deallocate(F->FilterString);
			F->FilterString = NFS;
			deallocate(DF);
		} else F->FilterString = DF;
	}
}

void swiFLAddOpenStrings(SWI_ID FilterID, char* Description, char* FilterString) {
	swiFLAddStrings(GlobalOpenFilter, Description, FilterString);
	swiFLAddStrings(FilterID, Description, FilterString);
}

char* swiFLFilterString(SWI_ID FilterID) {
	Filter* F = swiSFind(FilterID, FILTER);
	if (F) return F->FilterString;
	return 0;
}

void swiFLDefaultExtensionForSave(SWI_ID FilterID, char* Extension) {
	Filter* F = swiSFind(FilterID, FILTER);
	if (F) {
		if (Extension) {
			strncpy(F->DefaultExtension, Extension, 7);
			F->DefaultExtension[7] = 0;
		} else F->DefaultExtension[0] = 0;
	}
}

char* swiFLDefaultExtension(SWI_ID FilterID) {
	Filter* F = swiSFind(FilterID, FILTER);
	if (F) return F->DefaultExtension;
	return 0;
}

void swiFLDelete(SWI_ID FilterID) {
	Filter* F = swiSFind(FilterID, FILTER);
	if (F) {
		if (F->UsageCount > 0) F->DeleteWhenUseless = 1;
		else {
			swiSRemove(FilterID, FILTER);
			deallocate(F->FilterString);
			deallocate(F);
		}
	}
}

void swiFLUse(SWI_ID FilterID) {
	Filter* F = swiSFind(FilterID, FILTER);
	if (F) F->UsageCount++;
}

void swiFLUnuse(SWI_ID FilterID) {
	Filter* F = swiSFind(FilterID, FILTER);
	if (F) {
		F->UsageCount--;
		if ((F->UsageCount == 0) && F->DeleteWhenUseless)
			swiFLDelete(FilterID);
	}
}

static INDEXEDQUEUE IDs;

int SeekFilterUsers(void* Data) {
	SWI_ID* FoundFilter;
	if (((SWI_ID*)Data)[1] == DATA_CLASS) {
		FoundFilter = allocate(sizeof(SWI_ID));
		*FoundFilter = *((SWI_ID*)Data);
		EnqueueIntoIndexedQueue(&IDs, FoundFilter);
	}
	return 0;
}

SWI_ID swiFLFindOwnerOfFilterWithString(char* FilterString) {
	char* Location, *FilterFilter;
	SWI_ID* ID, Found;
	InitIndexedQueue(&IDs);
	swiSIterateWithCallback(&SeekFilterUsers);
	FilterFilter = FilterString + strlen(FilterString) + 1;
	while (NumberOfItemsInIndexedQueue(&IDs)) {
		ID = DequeueFromIndexedQueue(&IDs);
		Location = swiFLFilterString(swiDCFilter(*ID));
		while (*Location) {
			if (strcmp(Location, FilterString)) {
				Location += strlen(Location) + 1; /* over description */
				Location += strlen(Location) + 1; /* over filter */
			} else {
				Location += strlen(Location) + 1; /* over description */
				if (strcmp(Location, FilterFilter)) {
					Location += strlen(Location) + 1; /* over filter */
				} else {
					Found = *ID;
					deallocate(ID);
					while (NumberOfItemsInIndexedQueue(&IDs)) {
						deallocate(DequeueFromIndexedQueue(&IDs));
					}
					DeleteIndexedQueue(&IDs);
					return Found;
				}
			}
		}
		deallocate(ID);
	}
	DeleteIndexedQueue(&IDs);
	return 0;
}

int swiFLFindIndexOfFilterString(SWI_ID FilterID, char* FilterString) {
	int Index;
	char* Location, *FilterFilter;
	Location = swiFLFilterString(FilterID);
	FilterFilter = FilterString + strlen(FilterString) + 1;
	Index = 1;
	while (*Location) {
		if (strcmp(Location, FilterString)) {
			Location += strlen(Location) + 1;
			Location += strlen(Location) + 1;
			Index++;
		} else {
			Location += strlen(Location) + 1;
			if (strcmp(Location, FilterFilter)) {
				Location += strlen(Location) + 1;
				Index++;
			} else return Index;
		}
	}
	return 0;
}
