/********************************************************************

	swi_Storage.c

	Ismo Krkkinen

	Last changed: 12.11.1999

	Stores data into a binary tree and gives each an unique (within
	the tree) identifier.

********************************************************************/

#include "swi_Storage.h"
#include "memctrl.h"
#include "bintree.h"

typedef struct {
	SWI_ID ID;
	void* Pointer;
} CacheElement;

typedef struct {
	BINTREE BT;
	CacheElement Cache[LAST_TYPE];
	unsigned int PrevID;
	CRITICAL_SECTION CS;
} Storage;

typedef struct {
	SWI_ID ID;
	SWIDataTypes DT;
} SWIStructHeader;

static Storage S;
static int Initialized = 0;

int Compare(void* a, void* b, void* info) {
	int aval, bval;
	aval = *((SWI_ID*)a);
	bval = *((SWI_ID*)b);
	if (aval < bval) return -1;
	else if (aval > bval) return 1;
	else return 0;
}

void swiSInitialize() {
	int k;
	if (!Initialized) {
		InitBintree(&(S.BT), &Compare);
		Initialized = 1;
		for (k = 0; k < LAST_TYPE; k++) {
			S.Cache[k].ID = 0;
			S.Cache[k].Pointer = 0;
		}
		S.PrevID = 0;
		InitializeCriticalSection(&(S.CS));
	}
}

static void* swiSRemoveMinimum() {
	CacheElement* CE;
	SWI_ID* M = DeleteMinimumFromBintree(&(S.BT));
	if (M) {
		CE = &(S.Cache[((SWIStructHeader*)M)->DT]);
		if (*M == CE->ID) {
			CE->ID = 0;
			CE->Pointer = 0;
		}
	}
	return M;
}

void swiSFinalize() {
	void* foo;
	if (Initialized) {
		while (foo = swiSRemoveMinimum()) deallocate(foo);
		FreeBintree(&(S.BT));
		DeleteCriticalSection(&(S.CS));
		Initialized = 0;
	}
}

void swiSInsert(void* Item) {
	EnterCriticalSection(&(S.CS));
	do {
		S.PrevID = 84128931 * S.PrevID + 93385629;
		*((SWI_ID*)Item) = S.PrevID;
	} while (!S.PrevID || InsertToBintreeNoDuplicates(&(S.BT), Item, 0));
	LeaveCriticalSection(&(S.CS));
}

void* swiSFind(SWI_ID ID, SWIDataTypes Type) {
	void* Found;
	EnterCriticalSection(&(S.CS));
	if (S.Cache[Type].ID != ID) {
		S.Cache[Type].ID = ID;
		S.Cache[Type].Pointer = FindFromBintree(&(S.BT), &ID, 0);
	}
	Found = S.Cache[Type].Pointer;
	LeaveCriticalSection(&(S.CS));
	return Found;
}

void* swiSRemove(SWI_ID ID, SWIDataTypes Type) {
	void* Found;
	EnterCriticalSection(&(S.CS));
	if (S.Cache[Type].ID != ID) Found = FindFromBintree(&(S.BT), &ID, 0);
	else {
		Found = S.Cache[Type].Pointer;
		S.Cache[Type].ID = 0;
		S.Cache[Type].Pointer = 0;
	}
	DeleteNodeFromBintree(&(S.BT), &ID, 0);
	LeaveCriticalSection(&(S.CS));
	return Found;
}

void swiSIterateWithCallback(SWI_INT_PVOID CallbackFunc) {
	EnterCriticalSection(&(S.CS));
	IterateBintreeWithCallback(&(S.BT), 1, CallbackFunc);
	LeaveCriticalSection(&(S.CS));
}

SWIDataTypes swiSType(SWI_ID Object) {
	SWIStructHeader* SH;
	SWIDataTypes Found;
	EnterCriticalSection(&(S.CS));
	SH = FindFromBintree(&(S.BT), &Object, 0);
	if (SH) Found = SH->DT;
	else Found = 0;
	LeaveCriticalSection(&(S.CS));
	return Found;
}
